from __future__ import print_function
from collections import defaultdict
from struct import Struct
import numpy as np
import pandas as pd
import gzip
import sys
import os
from sce import sce
from scipy.io import mmread
from scipy.sparse import csr_matrix
def read_quants_bin(base_location, clipped=False, mtype="data", rmode="rust"):
    '''
    Read the quants Sparse Binary output of Alevin and generates a dataframe
    Parameters
    ----------
    base_location: string
        Path to the folder containing the output of the alevin run
    clipped: bool (default False)
        Clip off all zero rows and columns
    mtype: "[data(default), tier, var, mean]"
        Alevin's matrix type to load into memory
    '''
    if not os.path.isdir(base_location):
        print("{} is not a directory".format( base_location ))
        sys.exit(1)

    base_location = os.path.join(base_location, "alevin")
    
    print(base_location)
    if not os.path.exists(base_location):
        print("{} directory doesn't exist".format( base_location ))
        sys.exit(1)

    data_type = "f"
    if mtype == "data":
        quant_file = os.path.join(base_location, "quants_mat.gz")
    elif mtype == "tier":
        data_type = "B"
        quant_file = os.path.join(base_location, "quants_tier_mat.gz")
    elif mtype == "mean":
        quant_file = os.path.join(base_location, "quants_mean_mat.gz")
    elif mtype == "var":
        quant_file = os.path.join(base_location, "quants_var_mat.gz")
    else:
        print("wrong mtype:".format( mtype ))
        sys.exit(1)

    if not os.path.exists(quant_file):
        print("quant file {} doesn't exist".format( quant_file ))
        sys.exit(1)

    if mtype in ["mean", "var"]:
        cb_file = os.path.join(base_location, "quants_boot_rows.txt")
    else:
        cb_file = os.path.join(base_location, "quants_mat_rows.txt")

    if not os.path.exists(cb_file):
        print("quant file's index: {} doesn't exist".format( cb_file ))
        sys.exit(1)

    gene_file = os.path.join(base_location, "quants_mat_cols.txt")
    
    if not os.path.exists(gene_file):
        print("quant file's header: {} doesn't exist".format(gene_file))
        sys.exit(1)

    cb_names = pd.read_csv(cb_file, header=None)[0].values
    gene_names = pd.read_csv(gene_file, header=None)[0].values
    num_genes = len(gene_names)
    num_cbs = len(cb_names)
    num_entries = int(np.ceil(num_genes/8))

    if rmode == "rust":
        print("Using rust mode with {} rows and {} columns".format(num_cbs, num_genes))
        mat = sce.read_quants(quant_file, num_cbs, num_genes)
        umi_matrix = csr_matrix((mat[2], mat[1], mat[0]), shape=(num_cbs, num_genes)).todense()
    else:
        with gzip.open( quant_file ) as f:
            line_count = 0
            tot_umi_count = 0
            umi_matrix = []

            header_struct = Struct( "B" * num_entries)
            while True:
                line_count += 1
                if line_count%100 == 0:
                    print ("\r Done reading " + str(line_count) + " cells", end= "")
                    sys.stdout.flush()
                try:
                    num_exp_genes = 0
                    exp_counts = header_struct.unpack_from( f.read(header_struct.size) )
                    for exp_count in exp_counts:
                        num_exp_genes += bin(exp_count).count("1")

                    data_struct = Struct( data_type * num_exp_genes)
                    sparse_cell_counts_vec = list(data_struct.unpack_from( f.read(data_struct.size) ))[::-1]
                    cell_umi_counts = sum(sparse_cell_counts_vec)

                except:
                    print ("\nRead total " + str(line_count-1) + " cells")
                    print ("Found total " + str(tot_umi_count) + " reads")
                    break

                if cell_umi_counts > 0.0:
                    tot_umi_count += cell_umi_counts

                    cell_counts_vec = []
                    for exp_count in exp_counts:
                        for bit in format(exp_count, '08b'):
                            if len(cell_counts_vec) >= num_genes:
                                break

                            if bit == '0':
                                cell_counts_vec.append(0.0)
                            else:
                                abund = sparse_cell_counts_vec.pop()
                                cell_counts_vec.append(abund)

                    if len(sparse_cell_counts_vec) > 0:
                        print("Failure in consumption of data")
                        print("left with {} entry(ies)".format(len(sparse_cell_counts_vec)))
                    umi_matrix.append( cell_counts_vec )
                else:
                    print("Found a CB with no read count, something is wrong")
                    sys.exit(1)


    alv = pd.DataFrame(umi_matrix)
    alv.columns = gene_names
    alv.index = cb_names
    if clipped:
        alv = alv.loc[:, (alv != 0).any(axis=0)]

    return alv

def read_fry_bootstraps_bin(
    base_location,
    num_bootstraps,
):
    '''
    Read the bootstraps as EDS output of Alevin and generates a dictionary of 
    dataframes, one for each cell. The dimension of each dataframe is 
    num_bootstraps x num_genes
    Parameters
    ----------
    base_location: string
        Path to the folder containing the output of the alevin run
    num_bootstraps: int (default False)
        number of bootstraps
    density: "[sparse (default), dense ]" (not in use)
        Load sparse alevin output or dense output(<v0.14.0)
    mtype: "[data(default), tier, var, mean]" (not in use)
        Alevin's matrix type to load into memory

    '''
    if not os.path.isdir(base_location):
        print("{} is not a directory".format( base_location ))
        sys.exit(1)

    #base_location = os.path.join(base_location, "alevin")
    print(base_location)
    if not os.path.exists(base_location):
        print("{} directory doesn't exist".format( base_location ))
        sys.exit(1)

    data_type = "f"
    quant_file = os.path.join(base_location, "bootstraps.eds.gz")

    if not os.path.exists(quant_file):
        print("bootstrap file {} doesn't exist".format( quant_file ))
        sys.exit(1)
        
    cb_file = os.path.join(base_location, "barcodes.txt")
    if not os.path.exists(cb_file):
        print("quant file's index: {} doesn't exist".format( cb_file ))
        sys.exit(1)
    gene_file = os.path.join(base_location, "gene_names.txt")
    if not os.path.exists(gene_file):
        print("quant file's header: {} doesn't exist".format( gene_file))
        sys.exit(1)

    cb_names = pd.read_csv(cb_file, header=None, sep='\t')[1].values
    gene_names = pd.read_csv(gene_file, header=None)[0].values
    num_genes = len(gene_names)
    num_entries = int(np.ceil(num_genes/8))
    
    bootstrap_dict = {}
    cb_index = 0
    with gzip.open( quant_file ) as f:
        line_count = 0
        tot_umi_count = 0
        umi_matrix = []

        header_struct = Struct( "B" * num_entries)

        while True:
            line_count += 1
            if line_count%100 == 0:
                print ("\r Done reading " + str(line_count) + " cells", end= "")
                sys.stdout.flush()
                
            if line_count%num_bootstraps == 0:
                alv = pd.DataFrame(umi_matrix)
                alv.columns = gene_names
                bootstrap_dict[cb_names[cb_index]] = alv
                cb_index += 1
                umi_matrix = []
                
            try:
                num_exp_genes = 0
                exp_counts = header_struct.unpack_from( f.read(header_struct.size) )
                for exp_count in exp_counts:
                    num_exp_genes += bin(exp_count).count("1")

                data_struct = Struct( data_type * num_exp_genes)
                sparse_cell_counts_vec = list(data_struct.unpack_from( f.read(data_struct.size) ))[::-1]
                cell_umi_counts = sum(sparse_cell_counts_vec)

            except:
                print ("\nRead total " + str(line_count-1) + " cells")
                print ("Found total " + str(tot_umi_count) + " reads")
                break

            if cell_umi_counts > 0.0:
                tot_umi_count += cell_umi_counts

                cell_counts_vec = []
                for exp_count in exp_counts:
                    for bit in format(exp_count, '08b'):
                        if len(cell_counts_vec) >= num_genes:
                            break

                        if bit == '0':
                            cell_counts_vec.append(0.0)
                        else:
                            abund = sparse_cell_counts_vec.pop()
                            cell_counts_vec.append(abund)

                if len(sparse_cell_counts_vec) > 0:
                    print("Failure in consumption of data")
                    print("left with {} entry(ies)".format(len(sparse_cell_counts_vec)))
                umi_matrix.append( cell_counts_vec )
            else:
                print("Found a CB with no read count, something is wrong")
                sys.exit(1)
        
    #alv.index = cb_names
    return bootstrap_dict


def read_eq_bin( base_location ):
    '''
    Read the Eqclasses Binary output of Alevin and generates a dataframe

    Parameters
    ----------
    base_location: string
        Path to the folder containing the output of the alevin run
    '''
    base_location = os.path.join(base_location, "alevin")
    if not os.path.isdir(base_location):
        print("{} is not a directory".format( base_location ))
        sys.exit(1)

    eq_file = os.path.join(base_location, "cell_eq_mat.gz")
    if not os.path.exists(eq_file):
        print("eqclass file {} doesn't exist".format( eq_file ))
        sys.exit(1)

    order_file = os.path.join(base_location, "cell_eq_order.txt")
    if not os.path.exists(order_file):
        print("cell order file {} doesn't exist".format( order_file ))
        sys.exit(1)

    header_struct = Struct("Q"*2)
    with gzip.open( eq_file ) as f:
        count = 0
        read_count = 0
        umiCounts = defaultdict(lambda: defaultdict(int))
        while True:
            count += 1
            if count%100 == 0:
                print ("\r Done reading " + str(count) + " cells", end= "")
                sys.stdout.flush()
            try:
                bc, num_classes = header_struct.unpack_from( f.read(header_struct.size) )
            except:
                print ("\nRead total " + str(count-1) + " cells")
                print ("Found total " + str(read_count) + " reads")
                break

            if num_classes != 0:
                data_struct = Struct("I"*2*num_classes)
                data = data_struct.unpack_from( f.read(data_struct.size) )
                for i in range(num_classes):
                    eqId = data[i]
                    eqCount = data[i+num_classes]

                    read_count += eqCount
                    umiCounts[bc][eqId] += eqCount
            else:
                print("Found a CB with no read count, something is wrong")
                sys.exit(1)

    print ("making data frame")
    adf = pd.DataFrame(umiCounts).fillna(0)
    adf.columns = [x[0] for x in pd.read_csv( order_file, header=None).values]

    return adf

def read_arborescences(base_location):
    '''
    Read the quants Sparse Binary output of Alevin and generates a dataframe
    Parameters
    ----------
    base_location: string
        Path to the folder containing the output of the alevin run
    Returns:
        a mult-level dictionary with CB -> Gene -> Arborescence_length -> Frequency
    '''
    if not os.path.isdir(base_location):
        print("{} is not a directory".format( base_location ))
        sys.exit(1)

    base_location = os.path.join(base_location, "alevin")
    print(base_location)
    if not os.path.exists(base_location):
        print("{} directory doesn't exist".format( base_location ))
        sys.exit(1)

    arbo_file = os.path.join(base_location, "arborescence_dump.txt.gz")
    if not os.path.exists(arbo_file):
        print("quant file {} doesn't exist".format( quant_file ))
        sys.exit(1)

    cell_gene_arbo_data = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
    with gzip.open( arbo_file ) as f:
        cell_count = 0
        total_fragments = 0

        while True:
            cell_count += 1
            if cell_count % 10 == 0:
                print ("\r Done reading " + str(cell_count) + " cells", end= "")
                sys.stdout.flush()
                
            # each cell has a header line of CB name, number of expressed genes, 
            # total number of fragments
            try:
                (cell_name, num_exp_genes, num_fragments) = f.readline().decode().strip().split()
                total_fragments += int(num_fragments)
            except:
                print("\nRead Total {} cells w/ {} fragments".format(cell_count-1, total_fragments))
                break
            
            # each Cell is followed by gene with at least one mapped fragment
            # and has gene_id, nuumber_of_types_of arbos, 
            # [#frags in arbo, frequency of such arbo]+
            for _ in range(int(num_exp_genes)):
                (gid, num_arbo_types, *arbo_data) = f.readline().decode().strip().split()
                gid = int(gid)
                
                # iterating over each gene's arborescence data
                for i in range(int(num_arbo_types)):
                    arbo_index = 2*i
                    arbo_length = int(arbo_data[arbo_index])
                    arbo_frequency = int(arbo_data[arbo_index+1])
                    
                    cell_gene_arbo_data[cell_name][gid][arbo_length] += arbo_frequency

    return cell_gene_arbo_data

def read_bfh(base_location, t2g_file, retype="counts"):
    base_location = os.path.join(base_location, "alevin")
    if not os.path.isdir(base_location):
        print("{} is not a directory".format( base_location ))
        sys.exit(1)

    bfh_file = os.path.join(base_location, "bfh.txt")
    if not os.path.exists(bfh_file):
        print("bfh file {} doesn't exist".format( bfh_file ))
        sys.exit(1)

    if not os.path.exists(t2g_file):
        print("t2g file {} doesn't exist".format( t2g_file ))
        sys.exit(1)

    t2g = pd.read_csv(t2g_file, header=None, sep="\t").set_index(0).to_dict()[1]

    if retype == "counts":
        read_matrix = defaultdict(lambda : defaultdict(int))
    elif retype == "umis":
        read_matrix = defaultdict(lambda : defaultdict(lambda : defaultdict(int)))

    # read in Alevin
    with open( bfh_file ) as f:
        T = int(f.readline())
        C = int(f.readline())
        E = int(f.readline())

        tname = []
        bc_id_to_name = []

        for _ in range(T):
            tname.append(f.readline().strip())

        for _ in range(C):
            bc_id_to_name.append(f.readline().strip())

        for idx,line in enumerate(f):
            toks = line.strip().split()
            num_labels = int(toks[0])
            tot_num_reads = int(toks[num_labels+1])
            genes = set([])

            for txp in toks[1:num_labels+1]:
                genes.add(t2g[ tname[int(txp)] ])

            idx = num_labels+2
            num_bcs = int(toks[idx])
            read_validator = 0

            for _ in range(num_bcs):
                idx += 1
                bc_name = bc_id_to_name[ int(toks[idx]) ]
                idx += 1
                num_umi = int(toks[idx])
                num_reads = 0

                for _ in range(num_umi):
                    idx += 2
                    num_reads += int(toks[idx])
                    if retype == "umis":
                        read_matrix[bc_name][ tuple(sorted(toks[1:num_labels+1])) ][ toks[idx-1] ] += int(toks[idx])

                read_validator += num_reads
                if retype == "counts":
                    read_matrix[bc_name][ tuple(sorted(list(genes))) ] += num_reads

            if read_validator != tot_num_reads:
                print ("ERROR")

    return read_matrix

def read_tenx(base, version=2):
    '''
    Specify the path to the folder containing matrix.mtx file
    '''
    if version == 2:
        mat = mmread(os.path.join(base, "matrix.mtx")).toarray()

        genes_path = os.path.join(base, "genes.tsv")
        genes = pd.read_csv(genes_path, header=None)[0].values

        barcodes_path = os.path.join(base, "barcodes.tsv")
        barcodes = pd.read_csv(barcodes_path, header=None)[0].values
    elif version == 3:
        mat_file = os.path.join(base, "matrix.mtx.gz")
        with gzip.open(mat_file) as f:
            mat = mmread(f).toarray()

        genes_path = os.path.join(base, "features.tsv.gz")
        with gzip.open(genes_path) as f:
            genes = pd.read_csv(f, header=None)[0].values

        barcodes_path = os.path.join(base, "barcodes.tsv.gz")
        with gzip.open(barcodes_path) as f:
            barcodes = pd.read_csv(f, header=None)[0].values
    else:
        print("Wrong version")

    cr = pd.DataFrame(mat).T
    cr.index = [x.strip().split("-")[0] for x in barcodes]
    cr.columns = genes

    return cr

def read_umi_tools(infile):
    '''
    Specify the umi_tools count output file
    '''
    naive = pd.read_csv(infile, index_col=0, sep="\t")
    return naive.T

def read_umi_graph(base_location, out_location, kind="dot"):
    '''
    A function to read the per cell level UMI graph output from Alevin
    i.e. a file with name cel_umi_graphs.gz and dumps per cell level
    separate dot(viz) file
    '''
    if not os.path.isdir(base_location):
        print("{} directory doesn't exist".format(base_location))
        sys.exit(1)

    if not os.path.isdir(out_location):
        print("{} directory doesn't exist".format(out_location))
        sys.exit(1)

    base_location = os.path.join(base_location, "alevin")
    if not os.path.isdir(base_location):
        print("{} is not a directory".format(base_location))
        sys.exit(1)

    graph_file = os.path.join(base_location, "cell_umi_graphs.gz")
    if not os.path.exists(graph_file):
        print("graph file {} doesn't exist".format(graph_file))
        sys.exit(1)

    with gzip.open(graph_file) as file_handle:
        for cell_index,cell_graph in enumerate(file_handle):
            toks = cell_graph.decode().strip().split("\t")
            fname = toks[0].strip()
            if kind == "dot":
                write_dot(toks[1:], os.path.join(out_location, fname+".dot.gz"))
            print("\r processed {} cell".format(cell_index), end="")

def write_dot(toks, file_name):
    '''
    write per cell level dot file for each cell separately
    '''
    with gzip.open(file_name, 'wb') as f:
        cell_graph = "digraph {} {{".format(file_name)

        # adding vertices
        for vid in range(int(toks[0])):
            cell_graph += "\n" + str(vid)

        #adding edges
        for edge_group in toks[1:]:
            edges = edge_group.strip().split(",")
            cell_graph += "\n{} -> {{ ".format(edges[0])
            cell_graph += ' '.join(edges[1:])
            cell_graph += " }"

        cell_graph += "\n}"

        f.write(cell_graph.encode())



from collections import defaultdict
from itertools import combinations
import Dedup


def get_max_span_networks(gene_net):
    nodes = gene_net.nodes
    edges = {k: v for k, v in gene_net.edges.items()}
    num_networks = len(nodes)

    # breaking ties randomly and very bad implementation
    # basically resolve transitive cases
    for child, parent in edges.items():
        node = list(parent)[0]
        if len(parent) == 1 and node in edges and len(edges[node]) == 1:
            gene_net.edges[child] = gene_net.edges[node]
            del gene_net.edges[node]
            delIdx = gene_net.nodes.index(node)
            gene_net.nodes.pop(delIdx)
            gene_net.node_wts.pop(delIdx)
            num_networks -= 1

    # find leaves
    edges = {k: v for k, v in gene_net.edges.items()}
    internal_nodes = set([node for nodes in edges.values() for node in nodes])
    leaves = set(gene_net.edges) - internal_nodes

    # start DFS currently break ties randomly
    for leaf in leaves:
        curr = leaf
        while True:
            if curr in edges:
                # parents = edges[curr]
                parent = list(edges[curr])[0]
                del edges[curr]
                num_networks -= 1
                curr = parent
            else:
                break
    return num_networks


def rescue_UMI(eq_obj, min_set_txps, predictions):
    # first make new eqclass based on min_set_txps
    new_eqclasses = defaultdict(lambda: defaultdict(int))
    for i in range(eq_obj.num_eqclasses):
        try:
            labels = eq_obj.labels[i]
        except:
            print("ERROR: wrong Number of eqclasses provided")
            exit(1)
        umis = eq_obj.umis[i]

        new_label = []
        genes_set = set([])
        # add txp to new label only if in min set
        for label in labels:
            txp_name = eq_obj.txps_list[label]
            genes_set.add(eq_obj.txp_to_gene_dict[txp_name])
            if txp_name in min_set_txps:
                new_label.append(label)

        # Assign new label to new eqclass only if all txps were from one gene
        if len(genes_set) == 1:
            sorted_labels = tuple(sorted(new_label))
            for umi, count in umis.items():
                new_eqclasses[sorted_labels][umi] += count
        else:
            print("ERROR: {} Eqclass has txps from different gene", i)
            exit(1)
    print("New Alevin Eqclasses: ", new_eqclasses)

    # maintain a network of umis at each txp level for 1-length eqclass
    # Note No de-duplication yet
    txp_network = {}
    for labels, umis in new_eqclasses.items():
        if len(labels) == 1:
            txp_id = list(labels)[0]
            if txp_id in txp_network:
                print("ERROR: repetition of txp {} in new eqclasses".format(txp_id))
                exit(1)
            txp_network[txp_id] = Dedup.get_txp_umi_network(umis, txp_id)

    # making gene network
    gene_network = {}
    for gene in eq_obj.genes_list:
        txps = eq_obj.gene_to_txp_dict[gene]
        if gene in gene_network:
            print("ERROR: Duplicate gene network in Alevin_MST")
        net = Dedup.get_gene_umi_network(txps, txp_network)
        if net is not None:
            gene_network[gene] = net

    # Processing >1-length eqclasses
    for labels, umis in new_eqclasses.items():
        # Assuming all multi-g class are thrown away
        if len(labels) > 1:
            gene_name = eq_obj.txp_to_gene_dict[eq_obj.txps_list[labels[0]]]
            if gene_name in gene_network:
                Dedup.update_gene_network(umis, gene_network[gene_name], labels)
            else:
                # shouldn't be possible
                print("ERROR: SHOULDN'T REACH HERE (at least theoratically)")
                exit(1);

    for gene in eq_obj.genes_list:
        predictions[gene] = get_max_span_networks(gene_network[gene])


def check_set_coverage(eq_obj, txp_list):
    for eqclass in eq_obj.labels:
        covered = False
        for txp in eqclass:
            txp_name = eq_obj.txps_list[txp]
            if txp_name in txp_list:
                covered = True
                break
        if not covered:
            return False
    return True


def get_set_cover(eq_obj):
    # right_set is the set of txps
    right_set = eq_obj.txps_list

    # exhaustively search for all combination to get min_set cover
    grouping = 0
    min_txps = set([])
    while True:
        grouping = grouping + 1
        is_covered = False
        for txps in combinations(right_set, grouping):
            curr_min_txp_set = set(txps)
            is_covered = check_set_coverage(eq_obj, curr_min_txp_set)
            if is_covered:
                min_txps |= set(txps)
                break
        if is_covered:
            break
    print("min_set of txps for Alevin: ", min_txps)

    return min_txps


def get_prediction(eq_obj):
    predictions = {}
    for gene in eq_obj.genes_list:
        predictions[gene] = 0

    min_set_txps = get_set_cover(eq_obj)
    rescue_UMI(eq_obj, min_set_txps, predictions)

    return predictions



from collections import defaultdict

#change this function based on the use - case
# currently assuming u3 is 1-edit of u1 and u4
def one_distance(u1, u2):
    first = int(u1.replace("u", ""))
    second = int(u2.replace("u", ""))
    if abs(first-second) <= 1:
        return True
    else:
        return False

class Network:
    def __init__(self, umis, tid):
        self.nodes = list([uid+"_"+str(tid) for uid in umis.keys()])
        self.edges = defaultdict(set)
        self.node_wts = list(umis.values())
        self.populate_edges(umis, tid)

    def populate_edges(self, umis, tid):
        visit_list = sorted(umis, key=umis.get)

        while(len(visit_list) > 0):
            removed_umis = [ visit_list.pop() ]

            while len(removed_umis) != 0:
                query_umi = removed_umis.pop(0)
                query_freq = umis[query_umi]

                for umi in visit_list[::-1]:
                    umi_freq = umis[umi]
                    if one_distance(umi, query_umi)  and ((query_freq/2.0)+1)>umi_freq:
                        removed_umis.append(umi)

                for umi in removed_umis:
                    self.edges[umi+"_"+str(tid)].add( query_umi+"_"+str(tid) )
                    visit_list.remove(umi)
        print ("Edges for {} in Alevin_MST".format(tid))
        print (self.edges)

    def combine_nets(self, other_net):
        self.nodes += other_net.nodes
        self.node_wts += other_net.node_wts
        for k,v in other_net.edges.items():
            if k in self.edges:
                print ("ALEVIN MST: network can't be combined")
            self.edges[k] = v

    def one_distance(self, first_txp, second, txps):
        toks = first_txp.split("_")
        if len(toks) == 1:
            return one_distance(first_txp, second)
        else:
            return one_distance(toks[0], second)

def get_txp_umi_network(umis, txp):
    return Network(umis, txp)

def get_gene_umi_network(txps, txp_net):
    net = None
    for txp in txps:
        if txp in txp_net:
            other_net = txp_net[txp]
            if net is None:
                net = other_net
            else:
                net.combine_nets(other_net)
    return net

def update_gene_network(umis, gene_net, txps):
    nodes = gene_net.nodes
    # traverse over each umi in multi-t eqclass
    for umi, count in umis.items():
        # check if the umi is in 1-edit from any node in the graph
        for nid, node in enumerate(nodes):
            # if within 1-edit then add relevant edge
            if gene_net.one_distance(node, umi, set(txps)):
                node_freq = gene_net.node_wts[nid]
                if (node_freq/2.0)+1 > count:
                    gene_net.edges[umi].add(node)
                elif (count/2.0)+1 > node_freq:
                    gene_net.edges[node].add(umi)
    gene_net.nodes += umis.keys()
    gene_net.node_wts += umis.values()

def collapse(visit_list, umis, pcr_correct):
    removed_umis = [ visit_list.pop() ]
    num_collapsed = 0

    while len(removed_umis) != 0:
        query_umi = removed_umis.pop(0)
        query_freq = umis[query_umi]

        for umi in visit_list[::-1]:
            umi_freq = umis[umi]
            if one_distance(umi, query_umi):
                if not pcr_correct or \
                        ( pcr_correct and ((query_freq/2.0)+1)>umi_freq ):
                    removed_umis.append(umi)
                    num_collapsed += 1

        for umi in removed_umis:
            visit_list.remove(umi)

    return num_collapsed

def dedup(umis, pcr_correct):
    '''
    deduplicate based on the umi and it's frequency
    :param umis: dictionary of the UMI and it's count
    :return: counts after deduplicating
    '''
    sorted_umis_seqs = sorted(umis, key=umis.get)
    num_molecules = len(sorted_umis_seqs)

    while(len(sorted_umis_seqs) > 0):
        num_collapsed = collapse(sorted_umis_seqs, umis, pcr_correct)
        num_molecules = num_molecules - num_collapsed
        if num_molecules < 1:
            print("ERROR in UMI collapsing")

    return num_molecules



from collections import Counter
import numpy as np
import pandas as pd
import click


# example eq class
# 5 //number of transcripts
# 8 //numer of eqClasses
# t1 229 //list of txps and their lens
# t2 271
# t3 889
# t4 1022
# t5 1222
# 2 t1 t2 27 // #txps <\t> txpsIds_list_sep_by_<\t> <\t> count_of_reads
# ...

class Txps:
    def __init__(self, num_txps):
        self.T = num_txps
        self.txp_name = []
        self.txp_len = []
    
    def insert(self, name, length):
        '''
        insert element in txps object
        '''
        assert(len(self.txp_name) == len(self.txp_len)), "Wrong txp info insertion/deletion Done, Exiting"
        assert(len(self.txp_name) <= self.T), "More Txps than expected(" + str(self.T) +"), Exiting"

        self.txp_name.append(name)
        self.txp_len.append(length)
        
    def get_name(self, ind):
        '''
        return count of an eq class ind
        '''
        return self.txp_name[ind]
    
    def get_len(self, ind):
        '''
        return label of an eq class ind
        '''
        return self.txp_len[ind]

class EqClass:
    def __init__(self, num_eq_classes):
        self.eq_labels = []
        self.eq_count = []
        self.comb_wt = [0] * num_eq_classes
        self.E = num_eq_classes
        
    def get_num_classes(self):
        '''
        get count for the number of eq classes
        '''
        return self.E
        
    def insert(self, label, count, auxs, alphas, txp_obj):
        '''
        insert element in eqclass
        '''
        assert(len(self.eq_labels) == len(self.eq_count)), "Wrong EqClass insertion/deletion Done, Exiting"
        assert(len(self.eq_labels) <= self.E), "More Eq classes than expected(" + str(self.T) +"), Exiting"
        for x in label:
            assert(x < txp_obj.T), "Txp label "+ str(x) +" is wrong (more than max allowed) " + str(txp_obj.T-1) + ", Exiting"
        
        if isinstance(label, tuple):
            self.eq_labels.append(label)
        else:
            self.eq_labels.append(tuple(label))
        
        self.eq_count.append(count)
        
        if len(label) == 1:
            alphas[label[0]] += count
        
    def get_count(self, ind):
        '''
        return count of an eq class ind
        '''
        return self.eq_count[ind]
    
    def get_label(self, ind):
        '''
        return label of an eq class ind
        '''
        return self.eq_labels[ind]
    
    def get_comb_wt(self, ind):
        '''
        return label of an eq class ind
        '''
        return self.comb_wt[ind]
        
class Rld:
    def __init__(self, length_vector):
        self.R = len(length_vector)
        self.dist = Counter(length_vector)
        
def read_data(eq_file, rld_file):
    with open(eq_file) as f:
        # Number of txps
        T = int(f.readline())
        # Number of eq classes
        E = int(f.readline())
        
        # create txp object
        txp_obj = Txps(T)
        # create EqClass object
        eq_obj = EqClass(E)
        # alpahs based on unique counts
        alphas = [0.0] * T
        
        # Names of txps
        for _ in range(T):
            toks = f.readline().strip().split()
            txp_obj.insert(toks[0], int(toks[1]))

        # Parsing Eq Class
        for line in f:
            toks = line.strip().split()
            toks = map(int, toks)
            t = toks[0] # number of txps in this eq class
            eq_obj.insert(toks[1:t+1], int(toks[t+1]), toks[t+1:], alphas, txp_obj)
        
    with open(rld_file) as f:
        # create a rld object
        rld = f.readline().strip().split("\t")
        rld_obj = Rld(rld)
        
    return txp_obj, eq_obj, rld_obj, alphas
        

def populate_alphas(alphas_in, to, lengthcorrection):
    # increment every count by 0.5 to get initial starting point other than 0
    # from here https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L812
    alphas_in = np.array(alphas_in) + 0.5

    # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L834
    alphas_out = np.array([1.0] * to.T)

    # populate alphas
    for ind in range(to.T):
        if(lengthcorrection):
            eff_len = 1e-3 * to.get_len(ind)
        else:
            eff_len = 100
        # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L813
        alphas_in[ind] *= eff_len

    return alphas_out

def update_comb_wts(eo, to, ro, lengthcorrection):
    # No use of rich eq classes setting eqclass weight to 1
    # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L874
    # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L881
    # generating combined_weight for aux_prob (cnditional probability)
    for ind in range(eo.E):
        count = eo.get_count(ind)
        labels = eo.get_label(ind)

        norm = 0.0
        classLength = len(labels)
        comb_wts = np.array([])

        for lbl in labels:
            if(lengthcorrection):
                txp_len = float(to.get_len(lbl))
                comb_wt = count / txp_len
            else:
                comb_wt = count / classLength

            norm += comb_wt
            comb_wts = np.concatenate([ comb_wts, [comb_wt] ])

        # normalizing at per eq class level
        comb_wts /= norm
        eo.comb_wt[ind] = tuple(comb_wts)

def EMUpdate(to, eo, alphas_in, alphas_out):
    # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L240
    assert( len(alphas_in) == len(alphas_out) ), "size of alphas in and out does not match"
    
    for ind in range(eo.E):
        count = eo.get_count(ind)
        labels = eo.get_label(ind)
        auxs = eo.get_comb_wt(ind)
        
        gsize = len(labels)
        
        if (gsize > 1):
            denom = 0.0
            for i in range(gsize):
                tid = labels[i]
                aux = auxs[i]
                val = alphas_in[tid] * aux
                denom += val
#             assert( denom >= 0), "Denominator not summing up to at least zero for EMUpdate"
            
            if (denom > 0):
                invDenom = count / denom;
                for i in range(gsize):
                    tid = labels[i]
                    aux = auxs[i]
                    val = alphas_in[tid] * aux
                    if not np.isnan(val):
                        alphas_out[tid] += val * invDenom
        else:
            alphas_out[labels[0]] += count
        
def run_EM(to, eo, alphas_in, alphas_out):
    # skipping Eqclasses degeneracy test for now
    # ------------> {IMPORTANT} --------->
    # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L893
    # ------------> {IMPORTANT} --------->

    # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L900
    # A EM termination criteria, adopted from Bray et al. 2016
    minAlpha = 1e-8;
    alphaCheckCutoff = 1e-2;

    # https://github.com/COMBINE-lab/salmon/blob/a3bfcf72e85ebf8b10053767b8b506280a814d9e/src/SalmonQuantify.cpp#L2613
    maxIter = 10000
    relDiffTolerance = 0.01
    minIter = 50
    itNum = 0
    converged = False

    # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L917
    while (itNum < minIter) or (itNum < maxIter and not converged):
        # calling EMpdate function
        EMUpdate(to, eo, alphas_in, alphas_out)

        converged = True;
        maxRelDiff = -float("inf")

        # convergence criteria
        # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L525
        for ind in range(to.T):
            if alphas_out[ind] > alphaCheckCutoff:
                relDiff = abs(alphas_in[ind] - alphas_out[ind]) / alphas_out[ind]
                if relDiff > maxRelDiff:
                    maxRelDiff = relDiff

                if (relDiff > relDiffTolerance):
                    converged = False;

            alphas_in[ind] = alphas_out[ind]
            alphas_out[ind] = 0.0

        if itNum%100 == 0:
            print("iteration = "+str(itNum)+" | max rel diff. = " + str(maxRelDiff))
        itNum += 1

    print("iteration = "+str(itNum)+" | max rel diff. = " + str(maxRelDiff))

    # consistency checks
    # https://github.com/COMBINE-lab/salmon/blob/master/src/CollapsedEMOptimizer.cpp#L1014
    alpha_sum = 0.0
    for ind in range(to.T):
        alpha = alphas_in[ind]
        if alpha < minAlpha:
            alphas_in[ind] = 0.0
        else:
            alpha_sum += alpha
    assert(alpha_sum > 0), "Something wrong with salmon run, Exiting"

@click.command()
@click.option('--eq', help="path+name to input salmon format eqclass file")
@click.option('--out', help="path+name to output abundance file")
@click.option('--lengthcorrection', is_flag=True, default=False)
@click.option('--rld', help="path to file with fragment lengths")
def main(eq, out, lengthcorrection, rld):
    to, eo, ro, alphas_in = read_data(eq, rld)
    alphas_out = populate_alphas(alphas_in, to, lengthcorrection)
    update_comb_wts(eo, to, ro, lengthcorrection)
    run_EM(to, eo, alphas_in, alphas_out)
    with open(out, 'w') as f:
        f.write("Txp_Name\tCounts\n")
        for ind in range(to.T):
            f.write(to.get_name(ind)+"\t"+str(alphas_in[ind])+"\n")
            
if __name__ == "__main__":
    main()


# import click
from EqClass import Eqclasses
import Alevin
import Utools
import Kallisto
import CellRanger
import Alevin_MST
import sys

def plot_predictions(alv, utl, kal, crn, alv_mst):
    print("\n\n\n\n===================")
    print("IGNORE logs above this\n===================")
    print("Alevin Predictions: ")
    print(alv)
    print("Alevin_MST Predictions: ")
    print(alv_mst)
    print("Utools Predictions: ")
    print(utl)
    print("Kallisto Predictions: ")
    print(kal)
    print("CellRanger Predictions: ")
    print(crn)

def run(eq_file):
    '''
    Run tests for all the methods
    :param eq_file: path of the file having eqclass structure
    :return: None
    '''
    eq_class_obj = Eqclasses(eq_file)

    # get gene level prediction dictionary for each tool
    utl_prediction = Utools.get_prediction(eq_class_obj)
    alv_prediction = Alevin.get_prediction(eq_class_obj)
    alv_mst_prediction = Alevin_MST.get_prediction(eq_class_obj)
    kal_prediction = Kallisto.get_prediction(eq_class_obj)
    crn_prediction = CellRanger.get_prediction(eq_class_obj)

    # plot the histogram of the prediction for each tool
    plot_predictions(alv_prediction, utl_prediction,
                     kal_prediction, crn_prediction,
                     alv_mst_prediction)

if __name__ == "__main__":
    run(sys.argv[1])
